"""
   Copyright 2019-2021 Boris Shminke

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import os
import sqlite3
import subprocess
from argparse import ArgumentParser, Namespace
from functools import partial
from multiprocessing.pool import Pool
from typing import List, Optional, Tuple

import torch
from torch import Tensor
from tqdm import tqdm

from neural_semigroups import Magma
from neural_semigroups.utils import (
    connect_to_db,
    create_table_if_not_exists,
    hide_cells,
    insert_values_into_table,
    read_whole_file,
)

TABLE_NAME = "mace_output"


def write_mace_input(partial_table: Tensor, dim: int, filename: str) -> None:
    """
    write a randomised file in a Mace4 format

    :param partial_table: a Cayley table partially filled with :math:`-1`'s
    :param dim: total number of items in a magma
    :param filename: where to save the file
    :returns:
    """
    with open(filename, "w", encoding="utf-8") as in_file:
        print("formulas(assumptions).", file=in_file)
        print("(x * y) * z = x * (y * z).", file=in_file)
        for i in range(dim):
            for j in range(dim):
                cell = int(partial_table[i, j].item())
                if cell != -1:
                    print(f"{i} * {j} = {cell}.", file=in_file)
        print("end_of_list.", file=in_file)


def table_completion(
    dims: List[int], mace_timeout: int, mace_memory_mb: int, task_id: int
) -> Tuple[str, str]:
    """
    generate a random incomplete Cayley table and complete it

    :param dims: cardinalities of magmas to be generated in each task
    :param mace_timeout: number of seconds for ``mace4`` to search for a model or :math:``-1`` if it can search forever
    :param mace_memory_mb: number of memory used by a single ``mace4`` process in megabytes
    :param task_id: needed for using with multiprocessing
    :returns:
    """
    dim = dims[task_id]
    partial_table = hide_cells(
        Magma(cardinality=dim).cayley_table,
        int(torch.randint(1, dim * dim, (1,)).item()),
    )
    write_mace_input(partial_table, dim, f"{task_id}.in")
    with open(f"{task_id}.in", "r", encoding="utf-8") as task_in, open(
        f"{task_id}.out", "w", encoding="utf-8"
    ) as task_out, open(f"{task_id}.err", "w", encoding="utf-8") as task_err:
        subprocess.run(
            [
                "mace4",
                f"-n {dim}",
                f"-t {mace_timeout}",
                f"-b {mace_memory_mb}",
            ],
            stdin=task_in,
            stdout=task_out,
            stderr=task_err,
            check=False,
        )
    output = read_whole_file(f"{task_id}.out")
    errors = read_whole_file(f"{task_id}.err")
    for extension in ["in", "out", "err"]:
        os.remove(f"{task_id}.{extension}")
    return output, errors


def parse_args(args: Optional[List[str]]) -> Namespace:
    """

    :param args: a list of string arguments
        (for testing and use in a non script scenario)
    :returns: arguments namespace for the script
    """
    argument_parser = ArgumentParser()
    argument_parser.add_argument("--max_dim", type=int, required=True)
    argument_parser.add_argument("--min_dim", type=int, required=True)
    argument_parser.add_argument("--number_of_tasks", type=int, required=True)
    argument_parser.add_argument(
        "--number_of_processes", type=int, required=False, default=1
    )
    argument_parser.add_argument(
        "--mace_timeout", type=int, required=False, default=-1
    )
    argument_parser.add_argument(
        "--mace_memory_mb", type=int, required=False, default=500
    )
    argument_parser.add_argument("--database_name", type=str, required=True)
    parsed_args = argument_parser.parse_args(args)
    return parsed_args


def work_with_database(
    cursor: sqlite3.Cursor, args: Namespace, pool: Pool, dims: int
) -> None:
    """
    a function-helper

    :param cursor: a cursor for a database to work with
    :param args: additional arguments
    :param pool: a multiprocessing pool
    :param dims: semigroups cardinality
    :returns:
    """
    try:
        create_table_if_not_exists(
            cursor,
            TABLE_NAME,
            ["output STRING", "errors STRING"],
        )
        with tqdm(total=args.number_of_tasks) as progress_bar:
            for output, errors in pool.imap_unordered(
                partial(
                    table_completion,
                    dims,
                    args.mace_timeout,
                    args.mace_memory_mb,
                ),
                range(args.number_of_tasks),
            ):
                insert_values_into_table(cursor, TABLE_NAME, (output, errors))
                progress_bar.update()
    finally:
        pool.close()
        pool.join()


def generate_data_with_mace4(input_args: Optional[List[str]] = None) -> None:
    """
    :param input_args: a list of arguments
        (if ``None`` then ones from the command line are used)
    :returns:
    """
    args = parse_args(input_args)
    dims = (
        torch.randint(args.min_dim, args.max_dim + 1, (args.number_of_tasks,))
        .numpy()
        .tolist()
    )
    cursor = connect_to_db(args.database_name)
    pool = Pool(processes=args.number_of_processes)
    work_with_database(cursor, args, pool, dims)
    cursor.connection.close()



"""
   Copyright 2019-2021 Boris Shminke

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from typing import Callable, Optional

from torch.utils.data import TensorDataset


# pylint: disable=too-few-public-methods
class SemigroupsDataset(TensorDataset):
    """
    an extension of ``torch.util.data.TensorDataset``
    similar to a private class ``torchvision.datasets.vision.VisionDataset``
    """

    def __init__(
        self,
        root: str,
        cardinality: int,
        transform: Optional[Callable] = None,
    ):
        """
        :param root: root directory of dataset
        :param cardinality: a semigroup cardinality to use.
        :param transform: a function/transform that takes in a Cayley table
            and returns a transformed version.
        """
        super().__init__()
        self.root = root
        self.cardinality = cardinality
        self.transform = transform

    def __getitem__(self, index):
        tensors = super().__getitem__(index)
        if self.transform is not None:
            tensors = self.transform(tensors)
        return tensors



"""
   Copyright 2019-2021 Boris Shminke

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import gzip
import shutil
import sqlite3
from os import listdir
from os.path import getmtime, join, splitext
from typing import List, Tuple

import requests
import torch
from torch import Tensor
from tqdm import tqdm

from neural_semigroups.constants import CURRENT_DEVICE
from neural_semigroups.magma import Magma, get_two_indices_per_sample

# the Cayley table of Klein Vierergruppe
# pylint: disable=not-callable
FOUR_GROUP = torch.tensor(
    [[0, 1, 2, 3], [1, 0, 3, 2], [2, 3, 0, 1], [3, 2, 1, 0]]
)

# some non associative magma
# (0 * 1) * 2 = 0 * 2 = 0
# 0 * (1 * 2) = 0 * 0 = 1
# pylint: disable=not-callable
NON_ASSOCIATIVE_MAGMA = torch.tensor([[1, 0, 0], [2, 2, 0], [2, 2, 2]])


def random_semigroup(
    dim: int, maximal_tries: int
) -> Tuple[bool, torch.Tensor]:
    """
    randomly search for a semigroup Cayley table.
    Not recommended to use with dim > 4

    :param dim: number of elements in a semigroup
    :param maximal_tries: how many times to try at most
    :returns: a pair (whether the Cayley table is associative, a Cayley table of a magma)

    """
    associative = False
    try_count = 0
    magma = Magma(cardinality=dim)
    while not associative and try_count <= maximal_tries:
        magma = Magma(cardinality=dim)
        associative = magma.is_associative
        try_count += 1
    return associative, magma.cayley_table


def get_magma_by_index(cardinality: int, index: int) -> Magma:
    """
    find a magma from a lexicographical order by its index

    :param cardinality: the number of elements in a magma
    :param index: an index of magma in a lexicographical order
    :returns: a magma with a given index
    """
    square = cardinality ** 2
    if index < 0 or index >= cardinality ** square:
        raise ValueError(
            """
        An index must be non negative and less than $n^(n^2)$"""
        )
    cayley_table = []
    residual = index
    for _ in range(square):
        cayley_table.append(residual % cardinality)
        residual = residual // cardinality
    return Magma(
        torch.tensor(list(reversed(cayley_table))).reshape(
            cardinality, cardinality
        )
    )


def import_smallsemi_format(lines: List[str]) -> torch.Tensor:
    """
    imports lines in a format used by ``smallsemi`` `GAP package`.
    Format description:

    * filename is of a form ``data[n].gl``, :math:`1<=n<=7`
    * lines are separated by a pair of symbols ``\\r\\n``
    * there are exactly :math:`n^2` lines in a file
    * the first line is a header starting with '#' symbol
    * each line is a string of :math:`N` digits from :math:`0` to :math:`n-1`
    * :math:`N` is the number of semigroups in the database
    * each column represents a serialised Cayley table
    * the database contains only cells starting from the second
    * the first cell of each Cayley table is assumed to be filled with ``0``

    :param lines: lines read from a file of `smallsemi` format
    :returns: a list of Cayley tables

    .. _GAP package: https://www.gap-system.org/Manuals/pkg/smallsemi-0.6.12/doc/chap0.html

    """
    raw_tables = torch.tensor(
        [list(map(int, list(line[:-1]))) for line in lines[1:]]
    ).transpose(0, 1)
    tables = torch.cat(
        [torch.zeros([raw_tables.shape[0], 1], dtype=torch.long), raw_tables],
        dim=-1,
    )
    cardinality = int(tables.max()) + 1
    return tables.reshape(tables.shape[0], cardinality, cardinality)


def download_file_from_url(
    url: str, filename: str, buffer_size: int = 1024
) -> None:
    """
    downloads some file from the Web to a specified destination

    >>> import os
    >>> from neural_semigroups.constants import TEST_TEMP_DATA
    >>> temp_file = os.path.join(TEST_TEMP_DATA, "test.html")
    >>> if os.path.exists(temp_file):
    ...     os.remove(temp_file)
    >>> download_file_from_url("https://python.org/", temp_file)
    >>> os.path.exists(temp_file)
    True

    :param url: a valid HTTP URL
    :param filename: a valid filename
    :param buffer_size: a number of bytes to read from URL at once
    """
    response = requests.get(url, stream=True)
    file_size = int(response.headers.get("Content-Length", 0))
    progress = tqdm(
        response.iter_content(chunk_size=buffer_size),
        f"Downloading {filename}",
        total=int(file_size / buffer_size),
        unit="kB",
    )
    with open(filename, "wb") as file:
        for data in progress:
            file.write(data)
    response.close()


def find_substring_by_pattern(
    strings: List[str], starts_with: str, ends_before: str
) -> str:
    """
    search for a first occurrence of a given pattern in a string list

    >>> some_strings = ["one", "two", "three"]
    >>> find_substring_by_pattern(some_strings, "t", "o")
    'tw'
    >>> find_substring_by_pattern(some_strings, "four", "five")
    Traceback (most recent call last):
       ...
    ValueError: pattern four.*five not found

    :param strings: a list of strings where the pattern is searched for
    :param starts_with: the first letters of a pattern
    :param ends_before: a substring which marks the beginning of something different
    :returns: a pattern which starts with ``starts_with`` and ends before ``ends_before``
    """
    for package_name in strings:
        starting_index = package_name.find(starts_with)
        if starting_index >= 0:
            ending_index = package_name.find(ends_before)
            return package_name[starting_index:ending_index]
    raise ValueError(f"pattern {starts_with}.*{ends_before} not found")


def get_newest_file(dir_path: str) -> str:
    """
    get the last modified file from a diretory

    >>> from pathlib import Path
    >>> from os import makedirs, path
    >>> from neural_semigroups.constants import TEST_TEMP_DATA
    >>> shutil.rmtree(path.join(TEST_TEMP_DATA, "tmp"), ignore_errors=True)
    >>> makedirs(path.join(TEST_TEMP_DATA, "tmp"))
    >>> Path(path.join(TEST_TEMP_DATA, "tmp", "one")).touch()
    >>> from time import sleep
    >>> sleep(0.01)
    >>> Path(path.join(TEST_TEMP_DATA, "tmp", "two")).touch()
    >>> get_newest_file(path.join(TEST_TEMP_DATA, "tmp"))
    'test_temp_data/tmp/two'

    :param dir_path: a directory path
    :returns: the last modified file's name
    """
    return max(
        [join(dir_path, filename) for filename in listdir(dir_path)],
        key=getmtime,
    )


def make_discrete(cayley_cubes: torch.Tensor) -> torch.Tensor:
    """
    transforms a batch of probabilistic Cayley cubes and in the following way:

    * maximal probabilities in the last dimension become ones
    * all other probabilies become zeros

    >>> make_discrete(torch.tensor([
    ...    [[[0.9, 0.1], [0.1, 0.9]], [[0.8, 0.2], [0.2, 0.8]]],
    ...    [[[0.7, 0.3], [0.3, 0.7]], [[0.7, 0.3], [0.3, 0.7]]],
    ... ]))
    tensor([[[[1., 0.],
              [0., 1.]],
    <BLANKLINE>
             [[1., 0.],
              [0., 1.]]],
    <BLANKLINE>
    <BLANKLINE>
            [[[1., 0.],
              [0., 1.]],
    <BLANKLINE>
             [[1., 0.],
              [0., 1.]]]])

    :param cayley_cubes: a batch of probabilistic cubes representing Cayley tables
    :returns: a batch of probabilistic cubes filled in with ``0`` or ``1``
    """
    cardinality = cayley_cubes.shape[1]
    batch_size = cayley_cubes.shape[0]
    cayley_table = cayley_cubes.max(dim=-1)[1]
    cube = torch.zeros(
        [batch_size, cardinality, cardinality, cardinality],
        dtype=torch.float32,
    )
    sample_index, one, two = get_two_indices_per_sample(
        batch_size, cardinality
    )
    cube[sample_index, one, two, cayley_table[sample_index, one, two]] = 1.0
    return cube


def count_different(one: Tensor, two: Tensor) -> Tensor:
    """
    given two batches of the same size, counts number of positions in these
    batches, on which the tensor from the first batch differs from the second

    :param one: one batch of tensors
    :param two: another batch of tensors
    :returns: the number of different tensors
    """
    batch_size = one.shape[0]
    return (
        torch.zeros(batch_size)
        .where(
            torch.abs(one - two).reshape(batch_size, -1).max(dim=1)[0] > 0,
            torch.ones(batch_size),
        )
        .sum()
    )


def hide_cells(cayley_table: Tensor, number_of_cells: int) -> Tensor:
    """
    set several cells in a Cayley table to math:`-1`

    >>> torch.manual_seed(42) # doctest: +ELLIPSIS
    <torch...
    >>> hide_cells(torch.tensor([[0, 1], [2, 3]]), 2).cpu()
    tensor([[ 0,  1],
            [-1, -1]])

    :param cayley_table: a Cayley table
    :param number_of_cells: a number of cells to hide
    :returns: a Cayley table with hidden cells
    """
    partial_table = cayley_table.clone()
    cardinality = cayley_table.shape[0]
    pairs = torch.randperm(cardinality * cardinality)[:number_of_cells]
    partial_table[pairs // cardinality, pairs % cardinality] = -1
    return partial_table


def read_whole_file(filename: str) -> str:
    """
    reads the whole file into a string,  for example

    >>> read_whole_file("README.md").split("\\n")[2]
    '# Neural Semigroups'

    :param filename: a name of the file to read
    :returns: whole contents of the file
    """
    with open(filename, "r", encoding="utf-8") as input_file:
        text = input_file.read()
    return text


def partial_table_to_cube(table: Tensor) -> Tensor:
    """
    create a probabilistic cube from a partially filled Cayley table
    ``-1`` is translated to :math:`\\frac1n` where :math:`n` is table's
    cardinality, for example

    >>> partial_table_to_cube(torch.tensor([[0, -1], [0, 0]])).cpu()
    tensor([[[1.0000, 0.0000],
              [0.5000, 0.5000]],
    <BLANKLINE>
             [[1.0000, 0.0000],
              [1.0000, 0.0000]]])

    :param table: a Cayley table, partially filled by ``-1``'s
    :returns: a probabilistic cube
    """
    cardinality = table.shape[0]
    cube = torch.zeros(
        [cardinality, cardinality, cardinality],
        dtype=torch.float32,
        device=CURRENT_DEVICE,
    )
    rows, cols = torch.where(torch.ne(table, -1))
    cube[rows, cols, table[rows, cols]] = 1.0
    rows, cols = torch.where(torch.eq(table, -1))
    cube[rows, cols, :] = 1 / cardinality
    return cube.reshape([cardinality, cardinality, cardinality])


def create_table_if_not_exists(
    cursor: sqlite3.Cursor, table_name: str, columns: List[str]
) -> None:
    """
    create a table if it does not exist

    :param cursor: a cursor to the database where to create a table
    :param table_name: what table to create
    :param columns: a list of strings of format "COLUMN_NAME COLUMN_TYPE"
    :returns:
    """
    cursor.execute(
        "SELECT COUNT(*) FROM sqlite_master WHERE name = ?", [table_name]
    )
    if cursor.fetchone()[0] == 0:
        cursor.connection.execute(
            f"CREATE TABLE {table_name}({', '.join(columns)})"
        )


def insert_values_into_table(
    cursor: sqlite3.Cursor, table_name: str, values: Tuple[str, ...]
) -> None:
    """
    inserts values into a table

    :param cursor: a cursor to database where the target table is located
    :param table_name: the target table
    :param values: values to insert into the target table
    :returns:
    """
    placeholders = ", ".join(len(values) * ["?"])
    cursor.execute(f"INSERT INTO {table_name} VALUES({placeholders})", values)


def gunzip(archive_path: str) -> None:
    """
    extracts a GZIP file in the same folder

    :param archive_path: a path ending with ``.gz``
    :returns:
    """
    with gzip.open(archive_path, "rb") as f_in:
        with open(splitext(archive_path)[0], "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)


def connect_to_db(database_name: str) -> sqlite3.Cursor:
    """open a connection to an SQLite database

    :param database_name: filename of a database
    :returns: a cursor to the database
    """
    connection = sqlite3.connect(database_name, isolation_level=None)
    connection.execute("PRAGMA journal_mode=WAL;")
    return connection.cursor()



"""
   Copyright 2019-2021 Boris Shminke

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import re
import sqlite3
from typing import Callable, Optional

import torch
from tqdm import tqdm

from neural_semigroups.semigroups_dataset import SemigroupsDataset
from neural_semigroups.utils import connect_to_db


class Mace4Semigroups(SemigroupsDataset):
    """
    a ``torch.util.data.Dataset`` wrapper for the data of ``mace4`` output
    stored in a ``sqlite`` database

    >>> import shutil
    >>> from neural_semigroups.constants import TEST_TEMP_DATA
    >>> import os
    >>> from neural_semigroups.generate_data_with_mace4 import (
    ... generate_data_with_mace4)
    >>> shutil.rmtree(TEST_TEMP_DATA, ignore_errors=True)
    >>> os.mkdir(TEST_TEMP_DATA)
    >>> database = os.path.join(TEST_TEMP_DATA,"test.db")
    >>> torch.manual_seed(42) # doctest: +ELLIPSIS
    <torch...
    >>> generate_data_with_mace4([
    ... "--max_dim", "2",
    ... "--min_dim", "2",
    ... "--number_of_tasks", "1",
    ... "--database_name", database])
    >>> mace4_semigroups = Mace4Semigroups(
    ...     root=database,
    ...     cardinality=2,
    ...     transform=lambda x: x
    ... )
    >>> mace4_semigroups[0][0]
    tensor([[0, 0],
            [0, 0]])
    >>> mace4_semigroups.get_table_from_output("not a mace4 output file")
    Traceback (most recent call last):
        ...
    ValueError: wrong mace4 output file format!
    """

    _where_clause = "WHERE output LIKE '%Process % exit (max_models)%'"

    def __init__(
        self,
        cardinality: int,
        root: str,
        transform: Optional[Callable] = None,
    ):
        """
        :param root: a full path to an ``sqlite`` database file
           which has a table ``mace_output`` with a string column ``output``
        :param cardinality: the cardinality of semigroups
        :param transform: a function/transform that takes a Cayley table
            and returns a transformed version.
        """
        super().__init__(root, cardinality, transform)
        self.load_data_from_mace_output()

    def get_table_from_output(self, output: str) -> torch.Tensor:
        """
        gets a Cayley table of a magma from the output of ``mace4``

        :param output: output of ``mace4``
        :returns: a Cayley table
        """
        search_result = re.search(
            r".*function\(\*\(_,_\), \[(.*)]\)\..*", output, re.DOTALL
        )
        if search_result is None:
            raise ValueError("wrong mace4 output file format!")
        input_lines = search_result.groups()[0]
        # pylint: disable=not-callable
        cayley_table = torch.tensor(
            list(
                map(
                    int,
                    input_lines.translate(
                        str.maketrans("", "", " \t\n])")
                    ).split(","),
                )
            )
        ).view(self.cardinality, self.cardinality)
        return cayley_table

    def get_additional_info(self, cursor: sqlite3.Cursor) -> int:
        """
        gets some info from an SQLite database with ``mace4`` outputs

        :param cursor: an SQLite database cursor
        :returns: a total number of rows in a table, a magma dimension
        """
        cursor.execute(
            f"SELECT COUNT(*) FROM mace_output {self._where_clause}"
        )
        row_count = cursor.fetchone()[0]
        return row_count

    def load_data_from_mace_output(self) -> None:
        """loads data generated by ``mace4`` from an ``sqlite`` database"""
        cursor = connect_to_db(self.root)
        row_count = self.get_additional_info(cursor)
        cursor.execute(f"SELECT output FROM mace_output {self._where_clause}")
        features = []
        for _ in tqdm(range(row_count)):
            output = cursor.fetchone()[0]
            features.append(self.get_table_from_output(output))
        self.tensors = (torch.stack(features),)



"""
   Copyright 2019-2021 Boris Shminke

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import torch
from torch import Tensor
from torch.functional import einsum
from torch.nn import Module
from torch.nn.functional import kl_div

from neural_semigroups.utils import count_different, make_discrete


class AssociatorLoss(Module):
    """
    probabilistic associator loss
    """

    def __init__(self, discrete: bool = False):
        """
        :param discrete: when ``False``, the KL divergence is is used for measuring associativity in a continuous way.
            when ``True``, returns ``1`` for associative and ``0`` for non associative samples.
        """
        super().__init__()
        self.discrete = discrete

    def forward(self, cayley_cubes: Tensor) -> Tensor:
        """
        finds a probabilistic associator of a given probabilistic Cayley cube

        First, we build two 4-index tensors representing probability
        distributions of products :math:`a\\left(bc\\right)` and :math:`\\left(ab\\right)c`, respectively:

        :math:`T_{ijkl}=P\\left\\{e_i\\left(e_je_k\\right)=e_l\\right\\}=
        \\sum\\limits_{m=1}^nP\\left\\{e_ie_m=e_l\\vert e_je_k=e_m\\right\\}
        P\\left\\{e_je_k=e_m\\right\\}=\\sum\\limits_{m=1}^na_{iml}a_{jkm}`

        and

        :math:`T\\prime_{ijkl}=P\\left\\{\\left(e_ie_j\\right)e_k=e_l\\right\\}=
        \\sum\\limits_{m=1}^nP\\left\\{e_me_k=e_l\\vert e_ie_j=e_m\\right\\}
        P\\left\\{e_ie_j=e_m\\right\\}=\\sum\\limits_{m=1}^na_{mkl}a_{ijm}`

        Then we calculate `Kullback-Leibler divergence`_ between :math:`T_{ijkl}`
        and :math:`T\\prime_{ijkl}` to find a continuous measure of associativity of the input table.

        :param cayley_cubes: a batch of probabilistic Cayley cubes
        :returns: the probabilistic associator

        .. _Kullback-Leibler divergence: https://en.wikipedia.org/wiki/Kullback-Leibler_divergence

        """
        cubes = make_discrete(cayley_cubes) if self.discrete else cayley_cubes
        one = einsum("biml,bjkm->bijkl", cubes, cubes)
        two = einsum("bmkl,bijm->bijkl", cubes, cubes)
        batch_size = cayley_cubes.shape[0]
        associator = (
            count_different(one, two)
            if self.discrete
            else kl_div(torch.log(one), two, reduction="sum")
        )
        return associator / batch_size



"""
   Copyright 2019-2021 Boris Shminke

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from typing import Optional, Tuple

import torch


def get_two_indices_per_sample(
    batch_size: int, cardinality: int
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    generates all possible combination of two indices
    for each sample in a batch

    >>> get_two_indices_per_sample(1, 2)
    (tensor([0, 0, 0, 0]), tensor([0, 0, 1, 1]), tensor([0, 1, 0, 1]))

    :param batch_size: number of samples in a batch
    :param cardinality: number of possible values of an index
    :returns: triples (index, index, index)
    """
    a_range = torch.arange(cardinality)
    return (
        torch.arange(batch_size).repeat_interleave(cardinality * cardinality),
        a_range.repeat_interleave(cardinality).repeat(batch_size),
        a_range.repeat(cardinality).repeat(batch_size),
    )


class Magma:
    """
    an implementation of `a magma`_ (a set with a binary operation)

    .. _a magma: https://en.wikipedia.org/wiki/Magma_%28algebra%29
    """

    cayley_table: torch.Tensor

    def __init__(
        self,
        cayley_table: Optional[torch.Tensor] = None,
        cardinality: Optional[int] = None,
    ):
        """
        constucts a new magma

        >>> seed = torch.manual_seed(11)
        >>> Magma(cardinality=2)
        tensor([[1, 1],
                [0, 1]])
        >>> Magma(torch.tensor([[0, 1], [1, 0]]))
        tensor([[0, 1],
                [1, 0]])
        >>> Magma()
        Traceback (most recent call last):
            ...
        ValueError: at least one argument must be given
        >>> Magma(torch. tensor([[0]]), cardinality=2)
        Traceback (most recent call last):
            ...
        ValueError: cayley_table must be a `torch.Tensor` of shape (cardinality, cardinality)

        :param cayley_table: a Cayley table for a magma.
                             If not provided, a random table is generated.
        :param cardinality: a number of elements in a magma to generate a random one
        """
        if cayley_table is None:
            if cardinality is None:
                raise ValueError("at least one argument must be given")
            self.cayley_table = torch.randint(
                low=0,
                high=cardinality,
                size=[cardinality, cardinality],
                dtype=torch.long,
            )
        else:
            if cayley_table.shape[0] == cayley_table.shape[1] and (
                cardinality is None or cayley_table.shape[0] == cardinality
            ):
                self.cayley_table = cayley_table
            else:
                raise ValueError(
                    "cayley_table must be a `torch.Tensor` of shape "
                    "(cardinality, cardinality)"
                )

    def __repr__(self) -> str:
        return str(self.cayley_table)

    @property
    def cardinality(self) -> int:
        """
        number of elements in a magma
        """
        return self.cayley_table.shape[0]

    @property
    def is_associative(self) -> bool:
        """
        check associativity of a Cayley table

        :returns: whether the input table is associative or not

        """
        a_range = torch.arange(self.cardinality)
        square = self.cardinality ** 2
        one = a_range.repeat_interleave(square)
        two = a_range.repeat_interleave(self.cardinality).repeat(
            self.cardinality
        )
        three = a_range.repeat(square)
        return self.cayley_table[self.cayley_table[one, two], three].equal(
            self.cayley_table[one, self.cayley_table[two, three]]
        )

    @property
    def is_commutative(self) -> bool:
        """
        check commutativity of a Cayley table

        :returns: whether the input table is commutative or not

        """
        return torch.allclose(
            self.cayley_table, self.cayley_table.transpose(0, 1)
        )

    @property
    def identity(self) -> int:
        """
        find an identity element in a Cayley table

        :returns: the index of the identity element or -1 if there is no identity
        """
        a_range = torch.arange(self.cardinality)
        left_identity = (
            torch.eq(self.cayley_table, a_range).min(dim=1)[0].max(dim=0)
        )
        right_identity = (
            torch.eq(self.cayley_table.transpose(0, 1), a_range)
            .min(dim=1)[0]
            .max(dim=0)
        )
        if bool(
            left_identity[0]
            and right_identity[0]
            and left_identity[1].equal(right_identity[1])
        ):
            identity = int(left_identity[1])
        else:
            identity = -1
        return identity

    @property
    def has_inverses(self) -> bool:
        """
        check whether there are solutions of equations :math:`ax=b` and :math:`xa=b``
        """
        a_range = torch.arange(self.cardinality)
        return bool(
            torch.eq(self.cayley_table.sort()[0], a_range).min(dim=1)[0].min()
            and torch.eq(self.cayley_table.transpose(0, 1).sort()[0], a_range)
            .min(dim=1)[0]
            .min()
        )

    @property
    def probabilistic_cube(self) -> torch.Tensor:
        """
        a 3d array :math:`a` where :math:`a_{ijk}=P\\left\\{e_ie_j=e_k\\right\\}`

        :returns: a probabilistic cube representation of a Cayley table
        """
        cube = torch.zeros(
            [self.cardinality, self.cardinality, self.cardinality],
            dtype=torch.float32,
        )
        a_range = torch.arange(self.cardinality)
        one = a_range.repeat_interleave(self.cardinality)
        two = a_range.repeat(self.cardinality)
        cube[one, two, self.cayley_table[one, two]] = 1.0
        return cube

    @property
    def next_magma(self) -> "Magma":
        """
        goes to the next magma Cayley table in their lexicographical order

        >>> Magma(torch.tensor([[0, 1], [1, 0]])).next_magma
        tensor([[0, 1],
                [1, 1]])
        >>> Magma(torch.tensor([[0, 1], [1, 1]])).next_magma
        tensor([[1, 0],
                [0, 0]])
        >>> Magma(torch.tensor([[1, 1], [1, 1]])).next_magma
        Traceback (most recent call last):
            ...
        ValueError: there is no next magma!

        :returns: another magma
        """
        next_table = self.cayley_table.clone().detach()
        row, column = self.cardinality - 1, self.cardinality - 1
        while next_table[row, column] >= self.cardinality - 1:
            next_table[row, column] = 0
            row = row + (column - 1) // self.cardinality
            column = (column - 1) % self.cardinality
            if row < 0:
                raise ValueError("there is no next magma!")
        next_table[row, column] += 1
        return Magma(next_table)

    def random_isomorphism(self) -> torch.Tensor:
        """
        get some Cayley table isomorphic to ``self.cayley_table`` form example

        >>> Magma(torch.tensor([[0, 0], [0, 0]])).random_isomorphism()
        tensor([[1, 1],
                        [1, 1]])
        """
        permutation_tensor = torch.randperm(self.cayley_table.shape[0]).to(
            self.cayley_table.device
        )
        _, one, two = get_two_indices_per_sample(1, self.cayley_table.shape[0])
        isomorphic_cayley_table = torch.zeros(
            self.cayley_table.shape,
            dtype=torch.long,
            device=self.cayley_table.device,
        )
        isomorphic_cayley_table[
            permutation_tensor[one], permutation_tensor[two]
        ] = permutation_tensor[self.cayley_table[one, two]]
        return isomorphic_cayley_table



"""
   Copyright 2019-2021 Boris Shminke

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import torch
from torch import Tensor
from torch.nn import Module

from neural_semigroups.constants import CURRENT_DEVICE


class ConstantBaseline(Module):
    """
    A model that always fills in the same number
    """

    def __init__(self, cardinality: int, fill_in_with: int = 0):
        """
        initializes a constant model

        >>> ConstantBaseline(2, 3)
        Traceback (most recent call last):
            ...
        ValueError: `fill_in_with` should be non-negative and less than `cardinality`, got 3 >= 2
        >>> ConstantBaseline(2, 1).constant_distribution.cpu()
        tensor([0., 1.])

        :param cardinality: the number of elements in a magma
        :param fill_in_with: an item which will be suggested as a correct answer
        """
        super().__init__()
        if fill_in_with >= cardinality:
            raise ValueError(
                "`fill_in_with` should be non-negative "
                "and less than `cardinality`, "
                f"got {fill_in_with} >= {cardinality}"
            )
        self.cardinality = cardinality
        self.constant_distribution = torch.zeros(
            cardinality, dtype=torch.float
        ).to(CURRENT_DEVICE)
        self.constant_distribution[fill_in_with] = 1.0

    def forward(self, cayley_cube: Tensor) -> Tensor:
        """
        forward pass inhereted from Module

        >>> ConstantBaseline(2, 1)(torch.tensor([
        ...     [[[0., 1.], [0.5, 0.5]], [[1., 0.], [0., 1.]]],
        ...     [[[0., 1.], [1.0, 0.0]], [[0.5, 0.5], [0., 1.]]]
        ... ]).to(CURRENT_DEVICE)).cpu()
        tensor([[[[0., 1.],
          [0., 1.]],
        <BLANKLINE>
         [[1., 0.],
          [0., 1.]]],
        <BLANKLINE>
        <BLANKLINE>
        [[[0., 1.],
          [1., 0.]],
        <BLANKLINE>
         [[0., 1.],
          [0., 1.]]]])

        :param cayley_cube: probabilistic representation of a magma
        :returns: a batch of constant values (set in the constructor)
        """
        result = cayley_cube.clone().detach()
        result[cayley_cube.max(3)[0] != 1.0] = self.constant_distribution
        return result



"""
   Copyright 2019-2021 Boris Shminke

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from collections import OrderedDict
from typing import List, Tuple

import torch
from torch import Tensor
from torch.nn import BatchNorm1d, Linear, Module, ReLU, Sequential, Softmax2d

from neural_semigroups.constants import CURRENT_DEVICE


def get_linear_bn_relu_sequence(
    dims: List[int], layer_name_postfix: str
) -> Module:
    """
    constructs a sequential model of triples of
    linear, relu, and batch norm layers of given dimensions

    :param dims: dimensions for layers in a sequence
    :param layer_name_postfix:
    :returns: a sequential module
    """
    layers: "OrderedDict[str, Module]" = OrderedDict()
    for i in range(len(dims) - 1):
        layers.update(
            {
                f"linear{layer_name_postfix}{i}": Linear(
                    dims[i], dims[i + 1], bias=True
                )
            }
        )
        layers.update({f"relu{layer_name_postfix}{i}": ReLU()})
        layers.update({f"bn{layer_name_postfix}{i}": BatchNorm1d(dims[i + 1])})
    return Sequential(layers).to(CURRENT_DEVICE)


def get_encoder_and_decoder_layers(
    dims: List[int], split_last: bool
) -> Tuple[Module, Module]:
    """
    construct symmetrical encoder and decoder modules

    :param dims: the dimensions of layers in an encoder part (the same dimensions are used for a decoder)
    :param split_last: if ``True``, makes their last encoder dimensions twice larger than the first decoder dimension
        (for a reparametrization trick)
    :returns: a pair of two sequential models, representing encoder and decoder layers
    """
    encoder_dims = dims.copy()
    decoder_dims = list(reversed(dims))
    if split_last:
        encoder_dims[-1] *= 2
    encoder_layers = get_linear_bn_relu_sequence(encoder_dims, "0")
    decoder_layers = get_linear_bn_relu_sequence(decoder_dims, "1")
    return encoder_layers, decoder_layers


class MagmaDAE(Module):
    """
    Denoising Autoencoder for probability Cayley cubes of magmas
    """

    def __init__(
        self,
        cardinality: int,
        hidden_dims: List[int],
        do_reparametrization: bool = False,
    ):
        """
        :param cardinality: the number of elements in a magma
        :param hidden_dims: a list of sizes of hidden layers of the encoder and the decoder
        :param do_reparametrization: if ``True``, adds a reparametrization trick
        """
        super().__init__()
        self.cardinality = cardinality
        self.do_reparametrization = do_reparametrization
        (
            self.encoder_layers,
            self.decoder_layers,
        ) = get_encoder_and_decoder_layers(
            [cardinality ** 3] + hidden_dims, do_reparametrization
        )
        # pylint: disable=not-callable
        self._nearly_zero = torch.tensor([1e-6], device=CURRENT_DEVICE)
        # pylint: disable=not-callable
        self._nearly_one = torch.tensor(
            [1 - (self.cardinality - 1) * 1e-6], device=CURRENT_DEVICE
        )

    def encode(self, input_with_noise: Tensor) -> Tensor:
        """
        represent input cube as an embedding vector

        :param input_with_noise: a tensor with two indices
        :returns: some tensor with two indices and non-negative values
        """
        # pylint: disable=not-callable
        return self.encoder_layers(
            input_with_noise.view(input_with_noise.shape[0], -1)
        )

    def decode(self, encoded_input: Tensor) -> Tensor:
        """
        represent an embedding vector as something with size aligned with the
        input

        :param encoded_input: an embedding vector
        :returns: a vector of values from ``0`` to ``1`` (kind of probabilities)
        """
        # pylint: disable=not-callable
        return (
            Softmax2d()(
                self.decoder_layers(encoded_input)
                .view(-1, self.cardinality, self.cardinality, self.cardinality)
                .transpose(1, 3)
                .transpose(2, 3)
            )
            .transpose(2, 3)
            .transpose(1, 3)
        )

    def reparametrize(self, mu_and_log_sigma: Tensor) -> Tensor:
        """
        do a reparametrization trick

        :param mu_and_sigma: vector of expectation and standard deviation
        :returns: sample from a distribution
        """
        if self.do_reparametrization:
            dim = mu_and_log_sigma.shape[1] // 2
            sample = mu_and_log_sigma[:, :dim] + torch.exp(
                mu_and_log_sigma[:, dim:]
            ) * torch.randn(
                (mu_and_log_sigma.shape[0], dim), device=CURRENT_DEVICE
            )
        else:
            sample = mu_and_log_sigma
        return sample

    def forward(self, cayley_cubes: Tensor) -> Tensor:
        """
        forward pass inherited from Module

        :param cayley_cubes: a batch of probabilistic representations of magmas
        :returns: auto-encoded probabilistic representations of magmas
        """
        encoded_input = self.encode(cayley_cubes)
        reparametrized_input = self.reparametrize(encoded_input)
        decoded_input = self.decode(reparametrized_input)
        return torch.where(
            torch.eq(cayley_cubes, 1.0),
            self._nearly_one,
            torch.where(
                torch.eq(cayley_cubes, 0.0),
                self._nearly_zero,
                decoded_input,
            ),
        )
